import uvicorn
from typing import Optional, List, Dict, Any, Union
from pydantic import BaseModel, Field
import sys, json
from fastapi import FastAPI
from pathlib import Path
import gc
try:
    from smart_open import open
except:
    pass
try:
    import ctypes
    libc = ctypes.CDLL("libc.so.6")
    def free_memory():
        gc.collect()
        libc.malloc_trim(0)
except:
    def free_memory():
        gc.collect()
sys.path.append("../src")
from TRecSys.strategies import AvgUserStrategy
import pandas as pd

strategy = None
api = FastAPI()



class Column(BaseModel):
    field: str
    values: List[str]
    type: Optional[str]
    default: Optional[str]
    weight: Optional[float]
    window: Optional[List[float]]
    url: Optional[str]
    entity: Optional[str]
    environment: Optional[str]
    feature: Optional[str]
    length: Optional[int]


class Schema(BaseModel):
    encoders: List[Column]
    metric: Optional[str]='ip'
    filters: Optional[List[Column]]=[]
    user_encoders: Optional[List[Column]]=[]
    id_col: Optional[str]='id'
    
    def to_dict(self):
        return {
            "metric": self.metric,
            "filters": [vars(c) for c in self.filters],
            "encoders": [vars(c) for c in self.encoders],
        }


class KnnQuery(BaseModel):
    data: Dict[str, Union[List[str],str]]
    k: int
    explain:Optional[bool]=False

class KnnUserQuery(BaseModel):
    item_history: List[str]
    data: Optional[Dict[str, Union[List[str],str]]]={}
    k: int


@api.get("/")
async def read_root():
    return {"status": "OK", "schema_initialized": strategy.schema_initialized(), "total_items": strategy.get_total_items()}


@api.get("/partitions")
async def api_partitions():
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    ret = strategy.get_partition_stats()
    ret["status"] = "OK"
    return ret


@api.post("/fetch")
def api_fetch(lbls: List[str]):
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    if strategy.get_total_items()==0:
        return {"status": "error", "message": "No items are indexed"}
    return strategy.fetch(lbls)

@api.post("/encode")
async def api_encode(data: Dict[str, str]):
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    vec = strategy.encode(data)
    return {"status": "OK", "vec": [float(x) for x in vec]}


@api.post("/init_schema")
def init_schema(sr: Schema):
    schema_dict = sr.to_dict()
    partitions, enc_sizes = strategy.init_schema(**schema_dict)
    free_memory()
    return {"status": "OK", "partitions": len(partitions), "vector_size":strategy.get_embedding_dimension(), "feature_sizes":enc_sizes, "total_items":strategy.get_total_items()}

@api.post("/get_schema")
def get_schema():
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    else:
        return strategy.schema.to_dict()

@api.post("/index")
async def api_index(data: Union[List[Dict[str, str]], str]):
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    if type(data)==str:
        # read data remotely
        with open(data, 'r') as f:
            data = json.load(f)
    errors, affected_partitions = strategy.index(data)
    if any(errors):
        return {"status": "error", "items": errors}
    return {"status": "OK", "affected_partitions": affected_partitions}


@api.post("/query")
async def api_query(query: KnnQuery):
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    if strategy.get_total_items()==0:
        return {"status": "error", "message": "No items are indexed"}
    try:
        labels,distances, explanation =strategy.query(query.data, query.k, query.explain)
        if any(explanation):
            return {"status": "OK", "ids": labels, "distances": distances, "explanation":explanation}
        return {"status": "OK", "ids": labels, "distances": distances}
    except Exception as e:
        return {"status": "error", "message": str(e)}

@api.post("/user_query")
async def api_user_query(query: KnnUserQuery):
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    if strategy.get_total_items()==0:
        return {"status": "error", "message": "No items are indexed"}
    try:
        labels,distances =strategy.user_query(query.data, query.item_history, query.k)
        return {"status": "OK", "ids": labels, "distances": distances}
    except Exception as e:
        return {"status": "error", "message": str(e)}


@api.post("/save_model")
async def api_save(model_name:str):
    if not strategy.schema_initialized():
        return {"status": "error", "message": "Schema not initialized"}
    saved = strategy.save_model(model_name)
    return {"status": "OK", "saved_indices": saved}

@api.post("/load_model")
async def api_load(model_name:str):
    loaded = strategy.load_model(model_name)
    free_memory()
    return {"status": "OK", "loaded_indices": loaded}

@api.post("/list_models")
async def api_list():
    return strategy.list_models()

def run_server(config=None, host="0.0.0.0", port=5000, log_level="info"):
    global strategy
    strategy = AvgUserStrategy(config)
    uvicorn.run(api, host=host, port=port, log_level=log_level)

if __name__ == "__main__":
    from argparse import ArgumentParser
    argparse = ArgumentParser()
    argparse.add_argument('--host', default='0.0.0.0', type=str, help='host')
    argparse.add_argument('--port', default=5000, type=int, help='port')
    args = argparse.parse_args()
    
    data_dir = Path(__file__).absolute().parent.parent / "data"
    with (data_dir / "config.json").open('r') as f:
        config = json.load(f)
    strategy = AvgUserStrategy(config)
    uvicorn.run("__main__:api", host=args.host, port=args.port, log_level="info")



def get_values_nested(d):
    """
    get_values_nested ({'a':'a1','b':['b1', 'b2'],'c':{'c1':['c11','c12']}})
    returns: ['a1','b1', 'b2','c11','c12']
    """
    if type(d) == str:
        return [d]
    elif type(d) == list:
        return d
    elif type(d) == dict:
        return sum([get_values_nested(v) for v in d.values()], [])
    else:
        raise TypeError("Unsupported type " + str(type(d)))


def delistify_tree(tree, a, b):
    if type(tree) == str:
        return tree
    if type(tree) == list:
        return a if a in tree else (b if b in tree else None)
    if type(tree) == dict:
        ret = [(k, delistify_tree(v, a, b)) for k, v in tree.items()]
        return dict(filter(lambda x: x[1], ret))
    else:
        raise TypeError("Unsupported type " + str(type(d)))


def tree_find_depth(tree, a):
    if type(tree) == str:
        return 0 if tree == a else None
    if type(tree) == list:
        return 1 if a in tree else None
    if type(tree) == dict:
        try:
            found = next(filter(lambda v: v is not None, map(lambda v: tree_find_depth(v, a), tree.values())))
        except:
            return None
        return 1 + found
    else:
        raise TypeError("Unsupported type " + str(type(d)))


def are_siblings(tree, a, b):
    if type(tree) == str:
        return False
    if type(tree) == list:
        return a in tree and b in tree
    if type(tree) == dict:
        return any(are_siblings(v, a, b) for v in tree.values())
    else:
        raise TypeError("Unsupported type " + str(type(d)))


def lowest_depth(tree, a, b):
    if a == b:
        return 0
    if are_siblings(tree, a, b):
        return 1
    tree = delistify_tree(tree, a, b)
    print(tree_find_depth(tree, a))
    found_a = tree_find_depth(tree, a)
    found_b = tree_find_depth(tree, b)
    while found_a is not None and found_b is not None:
        for v in tree.values():
            da = tree_find_depth(v, a)
            db = tree_find_depth(v, b)
            if da is not None and db is not None:
                tree = v
                found_a = da
                found_b = db
                break
    return found_a + found_b



import numpy as np
import collections
try:
    import hnswlib
except ModuleNotFoundError:
    print ("hnswlib not found")
    HNSWMock = collections.namedtuple("HNSWMock", ("Index", "max_elements"))
    hnswlib = HNSWMock(None,0)
try:
    import faiss
except ModuleNotFoundError:
    print ("faiss not found")
    faiss = None

def parse_server_name(sname):
    if sname in ["hnswlib", "hnsw"]:
        return LazyHnsw
    elif sname in ["faiss", "flatfaiss"]:
        return FlatFaiss
    raise TypeError(str(sname) + " is not a valid similarity server name")

class FlatFaiss:
    def __init__(self, space, dim, **kwargs):
        if space in ['ip', 'cosine']:
            self.index = faiss.IndexFlatIP(dim)
            if space=='cosine':
                #TODO: Support cosine
                print("cosine is not supported yet, falling back to dot")
        elif space== 'l2':
            self.index = faiss.IndexFlatL2(dim)
        else:
            raise TypeError(str(space) + " is not supported")
        self.index=faiss.IndexIDMap2(self.index)

    def add_items(self, data, ids):
        data = np.array(data).astype(np.float32)
        self.index.add_with_ids(data, np.array(ids, dtype='int64'))

    def get_items(self, ids):
        # recmap = {k:i for i,k in enumerate(faiss.vector_to_array(self.index.id_map))}
        # return np.vstack([self.index.reconstruct(recmap[v]) for v in ids])
        return np.vstack([self.index.reconstruct(int(v)) for v in ids])

    def get_max_elements(self):
        return -1

    def get_current_count(self):
        return self.index.ntotal

    def search(self, data, k=1):
        return self.index.search(np.array(data).astype(np.float32),k)

    def save_index(self, fname):
        return faiss.write_index(self.index, fname)

    def load_index(self, fname):
        self.index = faiss.read_index(fname)

class LazyHnsw(hnswlib.Index):
    def __init__(self, space, dim, max_elements=1024, ef_construction=200, M=16):
        super().__init__(space, dim)
        self.init_max_elements = max_elements
        self.init_ef_construction = ef_construction
        self.init_M = M

    def init(self, max_elements=0):
        if max_elements == 0:
            max_elements = self.init_max_elements
        super().init_index(max_elements, self.init_M, self.init_ef_construction)

    def add_items(self, data, ids=None, num_threads=-1):
        if self.max_elements == 0:
            self.init()
        if self.max_elements<len(data)+self.element_count:
            super().resize_index(max([len(data)+self.element_count,self.max_elements]))
        return super().add_items(data, ids, num_threads)

    def add(self, data):
        return self.add_items(data)

    def get_items(self, ids=None):
        if self.max_elements == 0:
            return []
        return super().get_items(ids)

    def knn_query(self, data, k=1, num_threads=-1):
        if self.max_elements == 0:
            return [], []
        return super().knn_query(data, k, num_threads)

    def search(self, data, k=1):
        I,D = self.knn_query(data, k)
        return (D,I)

    def resize_index(self, size):
        if self.max_elements == 0:
            return self.init(size)
        else:
            return super().resize_index(size)

    def set_ef(self, ef):
        if self.max_elements == 0:
            self.init_ef_construction = ef
            return
        super().set_ef(ef)

    def get_max_elements(self):
        return self.max_elements

    def get_current_count(self):
        return self.element_count




import json, re, itertools, collections, os
from copy import deepcopy as clone
from operator import itemgetter as at
import numpy as np
from tree_helpers import lowest_depth, get_values_nested
import requests
from smart_open import open

class PartitionSchema:
    __slots__=["encoders", "filters", "partitions", "dim", "metric", "defaults", "id_col", "user_encoders"]
    def __init__(self, encoders, filters=[], metric='ip', id_col="id", user_encoders=[]):
        self.metric = metric
        self.id_col = id_col
        self.filters, self.partitions = self._parse_filters(filters)
        self.encoders = self._parse_encoders(encoders)
        self.user_encoders = self._parse_encoders(user_encoders)
        self.defaults = {}
        for f, e in self.encoders.items():
            if e.default is not None:
                self.defaults[f] = e.default
        self.dim = sum(map(len, filter(lambda e: e.column_weight!=0, self.encoders.values())))
    def _parse_encoders(self, encoders):
        encoder_dict = dict()
        for enc in encoders:
            if enc["type"] in ["onehot", "one_hot", "one hot", "oh"]:
                encoder_dict[enc["field"]] = OneHotEncoder(column=enc["field"], column_weight=enc["weight"],
                                                    values=enc["values"], default=enc.get("default"))
            elif enc["type"] in ["strictonehot", "strict_one_hot", "strict one hot", "soh"]:
                encoder_dict[enc["field"]] = StrictOneHotEncoder(column=enc["field"], column_weight=enc["weight"],
                                                            values=enc["values"], default=enc.get("default"))
            elif enc["type"] in ["num", "numeric"]:
                encoder_dict[enc["field"]] = NumericEncoder(column=enc["field"], column_weight=enc["weight"],
                                                    values=enc["values"], default=enc.get("default"))
            elif enc["type"] in ["ordinal", "ordered"]:
                encoder_dict[enc["field"]] = OrdinalEncoder(column=enc["field"], column_weight=enc["weight"],
                                                    values=enc["values"], default=enc.get("default"), window=enc["window"])
            elif enc["type"] in ["bin", "binning"]:
                encoder_dict[enc["field"]] = BinEncoder(column=enc["field"], column_weight=enc["weight"],
                    values=enc["values"], default=enc.get("default"))
            elif enc["type"] in ["binordinal", "bin_ordinal", "bin ordinal", "ord bin"]:
                encoder_dict[enc["field"]] = BinOrdinalEncoder(column=enc["field"], column_weight=enc["weight"],
                    values=enc["values"], default=enc.get("default"), window=enc["window"])
            elif enc["type"] in ["hier", "hierarchy", "nested"]:
                encoder_dict[enc["field"]] = HierarchyEncoder(column=enc["field"], column_weight=enc["weight"],
                                                        values=enc["values"], default=enc.get("default"),
                                                        similarity_by_depth=enc["similarity_by_depth"])
            elif enc["type"] in ["numpy", "np", "embedding"]:
                encoder_dict[enc["field"]] = NumpyEncoder(column=enc["field"], column_weight=enc["weight"],
                                                            values=enc["values"], default=enc.get("default"), url=enc["url"])
            elif enc["type"] in ["JSON", "json", "js"]:
                encoder_dict[enc["field"]] = JSONEncoder(column=enc["field"], column_weight=enc["weight"],
                                                            values=enc["values"], default=enc.get("default"), length=enc["length"])
            elif enc["type"] in ["qwak"]:
                encoder_dict[enc["field"]] = QwakEncoder(column=enc["field"], column_weight=enc["weight"],
                                                            length=enc["length"], entity_name=enc["entity"], default=enc.get("default"),
                                                            feature_name=enc["feature"], environment=enc["environment"])
            else:
                raise TypeError("Unknown type {t} in field {f}".format(f=enc["field"], t=enc["type"]))
        return encoder_dict
    def _parse_filters(self, filters):
        ret_filters = []
        partitions = [("ALL",)]
        if any(filters):
            partitions = list(itertools.product(*[f["values"] for f in filters]))
            ret_filters = [f["field"] for f in filters]
        return ret_filters, partitions
    def _unparse_encoders(self, encoders):
        return [dict({"field":k, "values": e.values, "type": type(e).__name__.replace("Encoder", "").lower(), "weight": e.column_weight, "default": e.default}, **e.special_properties()) for k, e in encoders.items()]
    def _unparse_filters(self, filters, partitions):
        ret_filters = collections.defaultdict(set)
        for i, k in enumerate(filters):
            for v in map(at(i), partitions):
                ret_filters[k].add(v)
        ret_filters = [{"field":k, "values": list(v)} for k,v in ret_filters.items()]
        return ret_filters
    def encode(self, x, weights=None):
        if type(x)==list:
            return np.vstack([self.encode(t,weights) for t in x])
        elif type(x)==dict:
            # Add default values to dict
            for f,d in self.defaults.items():
                if f not in x:
                    x[f] = d
            if weights is None:
                return np.concatenate([e(x[f]) for f, e in self.encoders.items() if e.column_weight!=0])
            assert len(weights)==len(self.encoders), "Invalid number of weight vector {w}".format(w=len(weights))
            return np.concatenate([w*e.encode(x[f])/np.sqrt(e.nonzero_elements) for f, e, w in zip(self.encoders.keys(),self.encoders.values(), weights)])
        else:
            raise TypeError("Usupported type for encode {t}".format(t=type(x)))
    def component_breakdown(self):
        start=0
        breakdown = {}
        for col,enc in self.encoders.items():
            if enc.column_weight==0:
                continue
            end = start + len(enc)
            breakdown[col] = (start, end)
        return breakdown
    def partition_num(self, x):
        if not any(self.filters):
            return 0
        combined_key = at(*(self.filters))(x)
        if type(combined_key)!=tuple:
            combined_key=(combined_key,)
        if all(type(k)!=list for k in combined_key):
            # one partition to lookup
            return self.partitions.index(combined_key)
        # multiple partitions
        combined_key_list = []
        for k in combined_key:
            if type(k)!=list:
                k=[k]
            combined_key_list.append(k)
        ret = []
        for combined_key in itertools.product(*combined_key_list):
            ret.append(self.partitions.index(combined_key))
        return ret
    def to_dict(self):
        filters = self._unparse_filters(self.filters, self.partitions)
        encoders = self._unparse_encoders(self.encoders)
        user_encoders = self._unparse_encoders(self.user_encoders)
        return {"encoders": encoders, "filters":filters, "metric": self.metric, "id_col": self.id_col, "user_encoders": user_encoders}


class BaseEncoder:

    def __init__(self, **kwargs):
        self.column = ''
        self.column_weight = 1
        self.nonzero_elements=1
        self.default = kwargs.get("default")
        self.__dict__.update(kwargs)

    def __len__(self):
        raise NotImplementedError("len is not implemented")

    def __call__(self, value):
        return self.column_weight * float(self.encode(value)) * np.ones(len(self)) * (1/np.sqrt(self.nonzero_elements))

    def encode(self, value):
        raise NotImplementedError("encode is not implemented")
    
    def special_properties(self):
        return {}

class CachingEncoder(BaseEncoder):
    cache_max_size=1024

    def __init__(self, **kwargs):
        # defaults:
        self.column = ''
        self.column_weight = 1
        self.values = []
        self.nonzero_elements=1
        self.default = kwargs.get("default")
        # override from kwargs
        self.__dict__.update(kwargs)
        #caching
        self.cache={}
        self.cache_hits=collections.defaultdict(int)

    def __call__(self, value):
        """Calls encode, multiplies by weight, cached"""
        if value in self.cache:
            self.cache_hits[value]+=1
            return self.cache[value]
        ret = self.encode(value)*self.column_weight*np.ones(len(self)) * (1/np.sqrt(self.nonzero_elements))
        if (self.cache_max_size is None) or (len(self.cache)<self.cache_max_size):
            self.cache[value]=ret
            return ret
        # cache cleanup
        min_key = sorted([(v,k) for k,v in self.cache_hits.items()])[0][1]
        del self.cache_hits[min_key]
        del self.cache[min_key]
        self.cache[value]=ret
        return ret

    def flush_cache(self,new_size=1024):
        self.cache_max_size=new_size
        self.cache={}
        self.cache_hits=collections.defaultdict(int)



class NumericEncoder(BaseEncoder):
    def __len__(self):
        return 1
    def encode(self, value):
        return np.array([float(value)])

class OneHotEncoder(CachingEncoder):

    def __len__(self):
        return len(self.values) + 1

    def encode(self, value):
        vec = np.zeros(1 + len(self.values))
        try:
            vec[1 + self.values.index(value)] = 1
        except ValueError:  # Unknown
            vec[0] = 1
        return vec

class StrictOneHotEncoder(CachingEncoder):
    def __len__(self):
        return len(self.values)

    def encode(self, value):
        vec = np.zeros(len(self.values))
        try:
            vec[self.values.index(value)] = 1
        except ValueError:  # Unknown
            pass
        return vec

class OrdinalEncoder(OneHotEncoder):
    def __init__(self, column, column_weight, values, window):
        super().__init__(column = column, column_weight=column_weight, values=values, window = window)
        self.window = window
        self.nonzero_elements=len(window)

    def encode(self, value):
        assert len(self.window) % 2 == 1, "Window size should be odd: window: {w}, value: {v}".format(w=self.window,v=value)
        vec = np.zeros(1 + len(self.values))
        try:
            ind = self.values.index(value)
        except ValueError:  # Unknown
            vec[0] = 1
            return vec
        vec[1 + ind] = self.window[len(self.window) // 2 + 1]
        for offset in range(len(self.window) // 2):
            if ind - offset >= 0:
                vec[1 + ind - offset] = self.window[len(self.window) // 2 - offset]
            if ind + offset < len(self.values):
                vec[1 + ind + offset] = self.window[len(self.window) // 2 + offset]

        return vec

    def special_properties(self):
        return {"widnow": self.window}


class BinEncoder(CachingEncoder):

    def __len__(self):
        return len(self.values) + 1

    def encode(self, value):
        value = float(value)
        vec = np.zeros(1 + len(self.values))
        i = 0
        while i < len(self.values) and value > self.values[i]:
            i += 1
        vec[i] = 1
        return vec


class BinOrdinalEncoder(BinEncoder):
    def __init__(self, column, column_weight, values, window):
        super().__init__(column = column, column_weight=column_weight, values=values, window = window)
        self.window = window
        self.nonzero_elements=len(window)

    def encode(self, value):
        value = float(value)
        vec = np.zeros(1 + len(self.values))
        ind = 0
        while ind < len(self.values) and value > self.values[ind]:
            ind += 1
        vec[ind] = self.window[len(self.window) // 2 + 1]
        for offset in range(len(self.window) // 2):
            if ind - offset >= 0:
                vec[ind - offset] = self.window[len(self.window) // 2 - offset]
            if ind + offset < len(self.values):
                vec[ind + offset] = self.window[len(self.window) // 2 + offset]
        return vec
    
    def special_properties(self):
        return {"window": self.window}


class HierarchyEncoder(CachingEncoder):
    #values = {'a': ['a1', 'a2'], 'b': ['b1', 'b2'], 'c': {'c1': ['c11', 'c12']}}
    similarity_by_depth = [1, 0.5, 0]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # TODO: Currently this approximation of the nonzero elements assumes:
        # (1) Two level hirearchy
        # (2) Approximately equal size of each hierarchy/category
        # {"A":[1,2,3], "B":[4,5,6]}
        inner_values = len(get_values_nested(self.values))
        outer_values = len(self.values.keys())
        self.nonzero_elements = inner_values / outer_values

    def __len__(self):
        inner_values = get_values_nested(self.values)
        return (1 + len(inner_values))

    def encode(self, value):
        # TODO: very inefficient: move to constructor
        inner_values = get_values_nested(self.values)
        vec = np.zeros(1 + len(inner_values))
        try:
            for other_value in inner_values:
                depth = lowest_depth(self.values, value, other_value)
                if depth >= len(self.similarity_by_depth):
                    # defaults to zero
                    continue
                vec[1 + inner_values.index(other_value)] = self.similarity_by_depth[depth]
        except ValueError:  # Unknown
            vec[0] = 1
        return vec

    def special_properties(self):
        return {"similarity_by_depth": self.similarity_by_depth}

class NumpyEncoder(BaseEncoder):
    def __init__(self, column, column_weight, values, url):
        super().__init__(column = column, column_weight=column_weight, values=values, url=url)
        with open(url, 'rb') as f:
            data = np.load(f)
            self.ids = list(data["ids"])
            self.embedding = data["embedding"]
        assert self.embedding.shape[0]==len(self.ids), "Dimension mismatch between ids and embedding"
        self.nonzero_elements=self.embedding.shape[1]

    def __len__(self):
        return self.embedding.shape[1]

    def encode(self, value):
        try:
            idx = self.ids.index(value)
        except ValueError:
            return np.zeros(self.embedding.shape[1])
        return self.embedding[idx,:]

class JSONEncoder(CachingEncoder):

    def __len__(self):
        return self.length

    def encode(self, value):
        val = value.translate({ord('('):'[',ord(')'):']'})
        val = json.loads(val)
        if type(val)==dict: # sparse vector
            vec = np.zeros(len(self))
            for k,v in val.items():
                vec[int(k)]=v
        elif type(val)==list:
            vec = np.array(val)
        else:
            raise TypeError(str(type(val))+ " is not supported")
        return vec

class QwakEncoder(BaseEncoder):
    def __init__(self, column, column_weight, environment, length, entity_name, api_key=None):
        super().__init__(column = column, column_weight=column_weight, length=length,
            entity_name=entity_name, environment=environment)
        if api_key is None:
            api_key = os.environ.get("QWAK_API")
        self.init_access_token(api_key)

    def init_access_token(self, api_key):
        self.access_token = requests.post("https://grpc.qwak.ai/api/v1/authentication/qwak-api-key", json={"qwakApiKey": api_key}).json()["accessToken"]

    def get_feature(self, entity_value):
        url = "https://api."+str(self.environment)+".qwak.ai/api/v1/features"
        body = {
            "features":[{"batchFeature":{"name": self.column}}],
            "entity": {"name": self.entity_name, "value": entity_value}
        }
        res = requests.post(url, headers={"Authorization": "Bearer "+self.access_token}, json=body).json()
        # Assuming one returned feature
        res = list(res["featureValues"][0]["featureValue"].values())[0]
        # TODO: Do qwak support vector features ?
        return res



    def __len__(self):
        return self.length

    def json_encode(self, value):
        val = value.translate({ord('('):'[',ord(')'):']'})
        val = json.loads(val)
        if type(val)==dict: # sparse vector
            vec = np.zeros(len(self))
            for k,v in val.items():
                vec[int(k)]=v
        elif type(val)==list:
            vec = np.array(val)
        else:
            raise TypeError(str(type(val))+ " is not supported")
        return vec

    def encode(self, value):
        val = self.get_feature(value)
        return self.json_encode(val)




import json, itertools, collections, sys
from operator import itemgetter as at
import numpy as np
from pathlib import Path
src = Path(__file__).absolute().parent
sys.path.append(str(src))
from encoders import PartitionSchema
from similarity_helpers import parse_server_name, LazyHnsw

class BaseStrategy:
    __slots__ = ["schema", "partitions","index_labels", "model_dir", "IndexEngine", "engine_params"]
    def __init__(self, config=None):
        if config is None:
            self.IndexEngine = LazyHnsw
            self.engine_params = {}
        else:
            self.engine_params=config[config["similarity_engine"]]
            self.IndexEngine = parse_server_name(config["similarity_engine"])
        self.model_dir = Path(__file__).absolute().parent.parent / "models"
        self.partitions = None
        self.schema =  None
        self.index_labels = []


    def init_schema(self, **kwargs):
        self.schema = PartitionSchema(**kwargs)
        self.partitions = [self.IndexEngine(self.schema.metric, self.schema.dim, **self.engine_params) for _ in self.schema.partitions]
        enc_sizes = {k:len(v) for k,v in self.schema.encoders.items()}
        return self.schema.partitions, enc_sizes

    def index(self, data):
        errors = []
        vecs = []
        for datum in data:
            try:
                vecs.append((self.schema.partition_num(datum), self.schema.encode(datum), datum[self.schema.id_col]))
            except KeyError as e:
                errors.append((datum, str(e)))
        vecs = sorted(vecs, key=at(0))
        affected_partitions = 0
        labels = set(self.index_labels)
        for partition_num, grp in itertools.groupby(vecs, at(0)):
            _, items, ids = zip(*grp)
            for id in ids:
                if id in labels:
                    errors.append((datum, "{id} already indexed.".format(id=id)))
                    continue # skip labels that already exists
                else:
                    labels.add(id)
                    self.index_labels.append(id)
            affected_partitions += 1
            num_ids = list(map(self.index_labels.index, ids))
            self.partitions[partition_num].add_items(items, num_ids)
        return errors, affected_partitions

    def query_by_partition_and_vector(self, partition_num, vec, k, explain=False):
        try:
            vec = vec.reshape(1,-1).astype('float32') # for faiss
            distances, num_ids = self.partitions[partition_num].search(vec, k=k)
        except Exception as e:
            raise Exception("Error in querying: " + str(e))
        if len(num_ids) == 0:
            labels, distances = [], []
        else:
            labels = [self.index_labels[n] for n in num_ids[0]]
            distances = [float(d) for d in distances[0]]
        if not explain:
            return labels,distances, []

        vec = vec.reshape(-1)
        explanation = []
        X = self.partitions[partition_num].get_items(num_ids[0])
        first_sim = None
        for ret_vec in X:
            start=0
            explanation.append({})
            for col,enc in self.schema.encoders.items():
                if enc.column_weight==0:
                    explanation[-1][col] = 0#float(enc.column_weight)
                    continue
                end = start + len(enc)
                ret_part = ret_vec[start:end]
                query_part =   vec[start:end]
                if self.schema.metric=='l2':
                    # The returned distance from the similarity server is not squared
                    dst=((ret_part-query_part)**2).sum()
                else:
                    sim=np.dot(ret_part,query_part)
                    # Correct dot product to be ascending
                    if first_sim is None:
                        first_sim = sim
                        dst = 0
                    else:
                        dst = 1-sim/first_sim
                explanation[-1][col]=float(dst)
                start = end
        return labels,distances, explanation

    def query(self, data, k, explain=False):
        try:
            partition_nums = self.schema.partition_num(data)
        except Exception as e:
            raise Exception("Error in partitioning: " + str(e))
        try:
            vec = self.schema.encode(data)
        except Exception as e:
            raise Exception("Error in encoding: " + str(e))
        if type(partition_nums)!=list:
            return self.query_by_partition_and_vector(partition_nums, vec, k, explain)
        # Aggregate results if multiple partitions are returned:
        labels,distances,explanation = [], [], []
        for partition_num in partition_nums:
            l,d,e = self.query_by_partition_and_vector(partition_num, vec, k, explain)
            labels.extend(l)
            distances.extend(d)
            #TODO: explanation is not supported when having multiple filters
        labels,distances = zip(*sorted(zip(labels,distances), key=at(1)))
        return labels,distances,explanation



    def save_model(self, model_name):
        (self.model_dir/model_name).mkdir(parents=True, exist_ok=True)
        with (self.model_dir/model_name/"index_labels.json").open('w') as f:
            json.dump(self.index_labels, f)
        with (self.model_dir/model_name/"schema.json").open('w') as f:
            json.dump(self.schema.to_dict(), f)
        saved=0
        for i,p in enumerate(self.partitions):
            fname = str(self.model_dir/model_name/str(i))
            try:
                p.save_index(fname)
                saved+=1
            except:
                continue
        return {"status": "OK", "saved_indices": saved}

    def load_model(self, model_name):
        with (self.model_dir/model_name/"schema.json").open('r') as f:
            schema_dict=json.load(f)
        schema = PartitionSchema(*schema_dict)
        partitions = [self.IndexEngine(schema.metric, schema.dim, **self.engine_params) for _ in self.partitions]
        (self.model_dir/model_name).mkdir(parents=True, exist_ok=True)
        with (self.model_dir/model_name/"index_labels.json").open('r') as f:
            self.index_labels=json.load(f)
        loaded = 0
        for i,p in enumerate(partitions):
            fname = str(self.model_dir/model_name/str(i))
            try:
                p.load_index(fname)
                loaded+=1
            except:
                continue
        return loaded, schema_dict

    def list_models(self):
        ret = [d.name for d in self.model_dir.iterdir() if d.is_dir()]
        ret.sort()
        return ret

    def fetch(self, lbls, numpy=False):
        sil = set(self.index_labels)
        found = [l for l in lbls if l in sil]
        ids = [self.index_labels.index(l) for l in found]
        ret = collections.defaultdict(list)
        for p,pn in zip(self.partitions, self.schema.partitions):
            for id in ids:
                try:
                    if numpy:
                        ret[pn].extend(p.get_items([id]))
                    else:
                        ret[pn].extend([tuple(float(v) for v in vec) for vec in p.get_items([id])])
                except Exception as e:
                    # not found
                    pass
        ret = map(lambda k,v: (k[0],v) if len(k)==1 else (str(k), v),ret.keys(), ret.values())
        ret = dict(filter(lambda kv: bool(kv[1]),ret))
        return ret

    def encode(self, data):
        return self.schema.encode(data)

    def schema_initialized(self):
        return (self.schema is not None)

    def get_partition_stats(self):
        display = lambda t: str(t[0]) if len(t)==1 else str(t)
        max_elements  = {display(p):self.partitions[i].get_max_elements()  for i,p in enumerate(self.get_partitions())}
        element_count = {display(p):self.partitions[i].get_current_count() for i,p in enumerate(self.get_partitions())}
        return {"max_elements": max_elements, "element_count":element_count, "n": len(self.partitions)}

    def get_partitions(self):
        return self.schema.partitions

    def get_embedding_dimension(self):
        return self.schema.dim

    def get_total_items(self):
        return len(self.index_labels)


class AvgUserStrategy(BaseStrategy):
    def user_partition_num(self, user_data):
        # Assumes same features as the item
        return self.schema.partition_num(user_data)

    def user_query(self, user_data, item_history, k, user_coldstart_item=None):
        if user_coldstart_item is None:
            n = 0
            vec = np.zeros(self.schema.dim)
        else:
            n = 1
            vec = self.schema.encode(user_coldstart_item)
        user_partition_num = self.user_partition_num(user_data)
        col_mapping = self.schema.component_breakdown()
        labels,distances = [], []
        if type(item_history) == str:
            item_history = [item_history]
        for item in item_history:
            # Calculate AVG
            item_vectors = [v for vs in self.fetch(item, numpy=True).values() for v in vs]
            n+=len(item_vectors)
            vec += np.sum(item_vectors, axis=0)
        if n>0:
            vec /= n
        # Override column values post aggregation, if needed
        for col, enc in self.schema.user_encoders.items():
            start, end = col_mapping[col]
            vec[start:end] = enc(user_data)
        # Query
        item_labels,item_distances,_ = self.query_by_partition_and_vector(user_partition_num, vec, k)
        labels.extend(item_labels)
        distances.extend(item_distances)
        return labels, distances




import json, collections, logging, os
import requests
import numpy as np
import pandas as pd
from operator import itemgetter as at
from tqdm import tqdm
from encoders import PartitionSchema
from pathlib import Path
from joblib import delayed, Parallel


def main(args):
	data_dir = Path(args.data_dir)
	with open(data_dir / "schema.json",'r') as f:
		schema = json.load(f)
	partition_dir=Path(args.partition_dir)
	partition_dir.mkdir(exist_ok=True)
	# copy schema
	logging.debug("Copy schema file")
	with (partition_dir/"schema.json").open('w') as f:
		json.dump(schema, f)
	#

	schema = PartitionSchema(schema)
	encode = schema.encode

	partitions = collections.defaultdict(pd.DataFrame)

	logging.debug("Repartition")
	p_files = [f for f in data_dir.glob("*.parquet")]
	for partition in tqdm(schema.partitions):
		kv=list(zip(schema.filters,partition))
		subindex = 0
		for f in p_files:
			df = pd.read_parquet(f)
			for k,v in kv:
				df=df[df[k]==v]
			n = len(df)//args.repartition_limit
			for sp in range(n+1):
				increment_df = df.iloc[:args.repartition_limit]
				if len(partitions[partition+(subindex,)])+len(increment_df)>args.repartition_limit:
					subindex+=1
				partitions[partition+(subindex,)]=pd.concat([partitions[partition+(subindex,)],increment_df])
				df=df.iloc[args.repartition_limit:]
				if len(increment_df)>0:
					subindex+=1

	logging.debug("Save to local json files")
	# save jsons
	j_files = []
	for partition,df in partitions.items():
		p = args.partition_sep.join(map(str,partition))
		j_file = partition_dir/(p)
		df.to_json(j_file, orient='records')
		j_files.append(j_file)

	logging.debug("Encoding each partition")
	def fstack(f,lst):
		"""Combines np.vstack and parallelism"""
		if len(lst)==0:
			return np.array([])
		elif len(lst)<1024:
			return np.vstack(Parallel(-1)([f(d) for d in lst]))
		return np.vstack([fstack(f,lst[:len(lst)//2]),fstack(f,lst[len(lst)//2:])])

	start_num_idx=0
	for jf in tqdm(j_files):
		partition = jf.name.split(args.partition_sep)
		subindex = int(partition[-1])
		partition = partition[:-1]
		with open(jf,'r') as f:
			data = json.load(f)
		arr = fstack(delayed(encode), data)
		np.save(jf.with_suffix('.npy'), arr)
		with jf.with_suffix('.meta').open('w') as f:
			index_num = schema.partition_num(dict(zip(schema.filters,partition)))
			meta = {
				"size":len(arr),
				"start_num_idx":start_num_idx,
				"ids": [d['id'] for d in data],
				"partition": partition,
				"subindex": subindex,
				"index_num":index_num,
				"metric":schema.metric,
				"dim":schema.dim,
				}
			json.dump(meta,f)
		start_num_idx+=len(arr)
		#jf.unlink(missing_ok=True) #TODO: why not working ?
		# os.remove(str(jf))



if __name__=="__main__":
    import sys
    from argparse import ArgumentParser
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
    argparse = ArgumentParser()
    argparse.add_argument('--data_dir',     default=str(Path(__file__).absolute().parent.parent / "data/ny")           ,type=str, help='data dir')
    argparse.add_argument('--partition_dir', default=str(Path(__file__).absolute().parent.parent / "data/ny/partitioned")   ,type=str, help='partition dir')
    argparse.add_argument('--partition_sep', default='~'   ,type=str, help='partition separator')
    argparse.add_argument('--repartition_limit', default=10000   ,type=int, help='repartition size limit')
    sys.exit(main(argparse.parse_args()))




from distutils.core import setup

__version__="0.2.4"
setup(
    name="TRecSys",
    packages=["TRecSys"],
    install_requires=[
        "faiss-cpu>=1.7.1.post3",
        "hnswlib>=0.5.1",
        "numpy>=1.21.2",
        "fastapi>=0.68.0",
        "uvicorn>=0.15.0",
        "smart_open[s3]~=3.0.0",
        "joblib>=0.17.0",
        "tqdm>=4.62.3",
        "pandas>=1.3.0",
    ],
    long_description="https://github.com/argmaxml/TRecs/blob/master/README.md",
    long_description_content_type="text/markdown",
    version=__version__,
    description='',
    author='ArgmaxML',
    author_email='uri@argmax.ml',
    url='https://github.com/argmaxml/TRecs',
    keywords=['recommendation-systems','recsys','similarity-server'],
    classifiers=[],
)



import json, logging, itertools
import requests
import numpy as np
import pandas as pd
from operator import itemgetter as at
from datetime import datetime
from pathlib import Path
from tqdm import tqdm
from joblib import delayed, Parallel
#from encoders import PartitionSchema
from similarity_helpers import parse_server_name

def index_one_partition(p, model_dir, sname, **sim_params):
    with p.with_suffix('.meta').open('r') as f:
        meta = json.load(f)
    Index = parse_server_name(sname)
    index_instance = Index(meta["metric"], meta["dim"], **sim_params)
    try:
        index_instance.load_index(str(model_dir/str(meta["index_num"])))
    except:
        pass
    arr=np.load(p)
    ids=np.arange(meta["start_num_idx"],meta["start_num_idx"]+meta["size"])
    if len(arr)==0:
        return []
    index_instance.add_items(arr,ids)
    index_instance.save_index(str(model_dir/str(meta["index_num"])))
    return [(int(i),str(l)) for i,l in zip(ids,meta["ids"])]


@delayed
def index_several_partition(paths, model_dir, sname, **sim_params):
    ret = []
    for p in paths:
        ret.extend(index_one_partition(p, model_dir, sname, **sim_params))
    return ret


def main(args):
    with Path(args.config_file).open('r') as f:
        config = json.load(f)
    Index = parse_server_name(config["similarity_engine"])
    sim_params=config[config["similarity_engine"]]
    partition_dir = Path(args.partition_dir)
    model_dir = Path(args.model_dir)
    model_dir.mkdir(exist_ok=True)
    logging.debug("Copy Schema")
    with (partition_dir/"schema.json").open('r') as i:
        with (model_dir/"schema.json").open('w') as o:
            json.dump(json.load(i),o)

    logging.debug("Start Index")
    start = datetime.now()
    #index_labels = Parallel(-1)([index_one_partition(p,model_dir,config["similarity_engine"],**config[config["similarity_engine"]]) for p in partition_dir.glob("*.npy")])
    partition_data = sorted([(p.name.rsplit(args.partition_sep,1)[0],p) for p in partition_dir.glob("*.npy")])
    partition_data = [[p for kk, p in grp] for k,grp in itertools.groupby(partition_data, key=at(0))]
    index_labels = Parallel(-1)([index_several_partition(ps,model_dir,config["similarity_engine"],**config[config["similarity_engine"]]) for ps in partition_data])
    index_labels = sorted(sum(index_labels,[]))
    index_labels = [l for i,l in index_labels]
    logging.debug("Save labels")
    with (model_dir/"index_labels.json").open('w') as f:
        json.dump(index_labels,f)

    end = datetime.now()
    logging.debug("Took {s} seconds to index".format(s=(end-start).seconds))
    return 0


if __name__=="__main__":
    import sys
    from argparse import ArgumentParser
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
    argparse = ArgumentParser()
    argparse.add_argument('--config_file',   default=str(Path(__file__).absolute().parent.parent / "data" / "config.json")  ,type=str, help='config file')
    argparse.add_argument('--model_dir',     default=str(Path(__file__).absolute().parent.parent / f"models/test-{datetime.now().strftime('%Y-%m-%d-%H-%M')}"), type=str, help='model dir')
    argparse.add_argument('--partition_dir', default=str(Path(__file__).absolute().parent.parent / "data/ny/partitioned")   ,type=str, help='partition dir')
    argparse.add_argument('--partition_sep', default='~'   ,type=str, help='partition separator')
    sys.exit(main(argparse.parse_args()))

